@@ -3,7 +3,7 @@
     Author:        Benoit Desouter <Benoit.Desouter@UGent.be>
                    Jan Wielemaker (SWI-Prolog port)
                    Fabrizio Riguzzi (mode directed tabling)
-    Copyright (c) 2016-2018, Benoit Desouter,
+    Copyright (c) 2016-2019, Benoit Desouter,
                              Jan Wielemaker,
                              Fabrizio Riguzzi
     All rights reserved.
@@ -61,6 +61,17 @@ table space and work lists are part of the SWI-Prolog core.
 @author Benoit Desouter, Jan Wielemaker and Fabrizio Riguzzi
 */
 
+% Enable debugging using debug(tabling(Topic)) when compiled with
+% -DO_DEBUG
+goal_expansion(tdebug(Topic, Fmt, Args), Expansion) :-
+    (   current_prolog_flag(prolog_debug, true)
+    ->  Expansion = debug(tabling(Topic), Fmt, Args)
+    ;   Expansion = true
+    ).
+% Simplify debugging of unexpected exceptions.  Assumes the code
+% should not raise exceptions.
+% goal_expansion(catch(Goal,_,_), Goal).
+
 %!  table(+PredicateIndicators)
 %
 %   Prepare the given PredicateIndicators for   tabling. Can only be
@@ -96,79 +107,89 @@ table(PIList) :-
 %           from future versions.
 
 start_tabling(Wrapper, Worker) :-
-    '$tbl_variant_table'(Wrapper, Trie, Status),
+    '$tbl_variant_table'(Wrapper, Trie, Status, Skeleton),
     (   Status == complete
-    ->  trie_gen(Trie, Wrapper, _)
-    ;   (   '$tbl_create_component'
-        ->  catch(run_leader(Wrapper, Worker, Trie),
-                  E, true),
-            (   var(E)
-            ->  trie_gen(Trie, Wrapper, _)
-            ;   '$tbl_table_discard_all',
-                throw(E)
-            )
-        ;   run_follower(Status, Wrapper, Worker, Trie)
-        )
+    ->  trie_gen(Trie, Skeleton, _)
+    ;   Status == fresh
+    ->  early_completion(Skeleton, Worker, Worker1),
+        '$tbl_create_subcomponent'(SCC),
+        setup_call_catcher_cleanup(
+            true,
+            run_leader(Skeleton, Worker1, Trie, SCC, LStatus),
+            Catcher,
+            finished_leader(Catcher, SCC, Wrapper)),
+        tdebug(schedule, 'Leader ~p done, status = ~p', [Wrapper, Status]),
+        done_leader(LStatus, SCC, Skeleton, Trie)
+    ;   % = run_follower, but never fresh and Status is a worklist
+        shift(call_info(Skeleton, Status))
     ).
 
-run_leader(Wrapper, Worker, Trie) :-
-    activate(Wrapper, Worker, Trie, _Worklist),
-    completion,
-    '$tbl_completed_component'.
+early_completion(v, Worker, once(Worker)) :- !.
+early_completion(_, Worker, Worker).
 
-run_follower(fresh, Wrapper, Worker, Trie) :-
+done_leader(complete, _SCC, Wrapper, Trie) :-
     !,
+    trie_gen(Trie, Wrapper, _).
+done_leader(final, SCC, Wrapper, Trie) :-
+    !,
+    '$tbl_free_component'(SCC),
+    trie_gen(Trie, Wrapper, _).
+done_leader(_,_,_,_).
+
+finished_leader(exit, _, _) :-
+    !.
+finished_leader(fail, _, _) :-
+    !.
+finished_leader(Catcher, SCC, Wrapper) :-
+    '$tbl_table_discard_all'(SCC),
+    (   Catcher = exception(_)
+    ->  true
+    ;   print_message(error, tabling(unexpected_result(Wrapper, Catcher)))
+    ).
+
+%!  run_leader(+Wrapper, +Worker, +Trie, +SCC, -Status) is det.
+%
+%   Run the leader of  a  (new)   SCC,  storing  instantiated  copies of
+%   Wrapper into Trie. Status  is  the  status   of  the  SCC  when this
+%   predicate terminates. It is one of   `complete`, in which case local
+%   completion finished or `merged` if running   the completion finds an
+%   open (not completed) active goal that resides in a parent component.
+%   In this case, this SCC has been merged with this parent.
+%
+%   If the SCC is merged, the answers   it already gathered are added to
+%   the worklist and we shift  (suspend),   turning  our  leader into an
+%   internal node for the upper SCC.
+
+run_leader(Wrapper, Worker, Trie, SCC, Status) :-
+    tdebug(schedule, '-> Activate component ~p for ~p', [SCC, Wrapper]),
     activate(Wrapper, Worker, Trie, Worklist),
-    shift(call_info(Wrapper, Worklist)).
-run_follower(Worklist, Wrapper, _Worker, _Trie) :-
-    shift(call_info(Wrapper, Worklist)).
+    tdebug(schedule, '-> Complete component ~p for ~p', [SCC, Wrapper]),
+    completion(SCC),
+    tdebug(schedule, '-> Completed component ~p for ~p', [SCC, Wrapper]),
+    '$tbl_component_status'(SCC, Status),
+    (   Status == merged
+    ->  tdebug(schedule, 'Turning leader ~p into follower', [Wrapper]),
+        (   trie_gen(Trie, Wrapper1, _),
+            tdebug(answer, 'Adding old answer ~p to worklist ~p',
+                   [ Wrapper1, Worklist]),
+            '$tbl_wkl_add_answer'(Worklist, Wrapper1),
+            fail
+        ;   true
+        ),
+        shift(call_info(Wrapper, Worklist))
+    ;   true                                    % completed
+    ).
 
 activate(Wrapper, Worker, Trie, WorkList) :-
     '$tbl_new_worklist'(WorkList, Trie),
+    %assertz(user:wl(Wrapper, WorkList)),
+    tdebug(activate, '~p: created wl=~p, trie=~p',
+           [Wrapper, WorkList, Trie]),
     (   delim(Wrapper, Worker, WorkList),
         fail
     ;   true
     ).
 
-
-%!  start_tabling(:Wrapper, :Implementation, +Variant, +ModeArgs)
-%
-%   As start_tabling/2, but in addition separates the data stored in the
-%   answer trie in the Variant and ModeArgs.
-
-start_tabling(Wrapper, Worker, WrapperNoModes, ModeArgs) :-
-    '$tbl_variant_table'(WrapperNoModes, Trie, Status),
-    (   Status == complete
-    ->  trie_gen(Trie, WrapperNoModes, ModeArgs)
-    ;   (   Status == fresh
-        ->  '$tbl_create_subcomponent',
-            catch(run_leader(Wrapper, WrapperNoModes, ModeArgs, Worker, Trie),
-                  E, true),
-            (   var(E)
-            ->  trie_gen(Trie, WrapperNoModes, ModeArgs)
-            ;   '$tbl_table_discard_all',
-                throw(E)
-            )
-        ;   % = run_follower, but never fresh and Status is a worklist
-            shift(call_info(Wrapper, Status))
-        )
-    ).
-
-get_wrapper_no_mode_args(M:Wrapper, M:WrapperNoModes, ModeArgs) :-
-    M:'$table_mode'(Wrapper, WrapperNoModes, ModeArgs).
-
-run_leader(Wrapper, WrapperNoModes, ModeArgs, Worker, Trie) :-
-    activate(Wrapper, WrapperNoModes, ModeArgs, Worker, Trie, _Worklist),
-    completion,
-    '$tbl_completed_component'.
-
-activate(Wrapper, WrapperNoModes, _ModeArgs, Worker, Trie, WorkList) :-
-    '$tbl_new_worklist'(WorkList, Trie),
-    (   delim(Wrapper, WrapperNoModes, Worker, WorkList),
-        fail
-    ;   true
-    ).
-
 %!  delim(+Wrapper, +Worker, +WorkList)
 %
 %   Call/resume Worker for non-mode directed tabled predicates.
@@ -181,6 +202,7 @@ delim(Wrapper, Worker, WorkList) :-
 
 work_and_add_answer(Worker, Wrapper, WorkList) :-
     call(Worker),
+    tdebug(answer, 'New answer ~p for ~p', [Wrapper,WorkList]),
     '$tbl_wkl_add_answer'(WorkList, Wrapper).
 
 
@@ -188,10 +210,55 @@ add_answer_or_suspend(0, _Wrapper, _WorkList, _) :-
     !.
 add_answer_or_suspend(Continuation, Wrapper, WorkList,
                       call_info(SrcWrapper, SourceWL)) :-
+    tdebug(schedule, 'Suspended ~p, for solving ~p', [SrcWrapper, Wrapper]),
     '$tbl_wkl_add_suspension'(
         SourceWL,
         dependency(SrcWrapper, Continuation, Wrapper, WorkList)).
 
+
+%!  start_tabling(:Wrapper, :Implementation, +Variant, +ModeArgs)
+%
+%   As start_tabling/2, but in addition separates the data stored in the
+%   answer trie in the Variant and ModeArgs.
+
+start_tabling(Wrapper, Worker, WrapperNoModes, ModeArgs) :-
+    '$tbl_variant_table'(WrapperNoModes, Trie, Status, _Skeleton),
+    (   Status == complete
+    ->  trie_gen(Trie, WrapperNoModes, ModeArgs)
+    ;   Status == fresh
+    ->  '$tbl_create_subcomponent'(SubComponent),
+        setup_call_catcher_cleanup(
+            true,
+            run_leader(Wrapper, WrapperNoModes, ModeArgs,
+                       Worker, Trie, SubComponent),
+            Catcher,
+            finished_leader(Catcher, SubComponent, Wrapper)),
+        tdebug(schedule, 'Leader ~p done, modeargs = ~p', [Wrapper, ModeArgs]),
+        trie_gen(Trie, WrapperNoModes, ModeArgs)
+    ;   % = run_follower, but never fresh and Status is a worklist
+        shift(call_info(Wrapper, Status))
+    ).
+
+get_wrapper_no_mode_args(M:Wrapper, M:WrapperNoModes, ModeArgs) :-
+    M:'$table_mode'(Wrapper, WrapperNoModes, ModeArgs).
+
+run_leader(Wrapper, WrapperNoModes, ModeArgs, Worker, Trie, SCC) :-
+    activate(Wrapper, WrapperNoModes, ModeArgs, Worker, Trie, _Worklist),
+    completion(SCC),
+    '$tbl_component_status'(SCC, Status),
+    (   Status == final
+    ->  '$tbl_free_component'(SCC)
+    ;   true                                    % What if merged?
+    ).
+
+
+activate(Wrapper, WrapperNoModes, _ModeArgs, Worker, Trie, WorkList) :-
+    '$tbl_new_worklist'(WorkList, Trie),
+    (   delim(Wrapper, WrapperNoModes, Worker, WorkList),
+        fail
+    ;   true
+    ).
+
 %!  delim(+Wrapper, +WrapperNoModes, +Worker, +WorkList).
 %
 %   Call/resume Worker for mode directed tabled predicates.
@@ -232,17 +299,28 @@ update(M:Wrapper, A1, A2, A3) :-
     A1 \=@= A3.
 
 
-%!  completion
+%!  completion(+Component)
 %
 %   Wakeup suspended goals until no new answers are generated.
 
-completion :-
-    '$tbl_pop_worklist'(WorkList),
-    !,
-    completion_step(WorkList),
-    completion.
-completion :-
-    '$tbl_table_complete_all'.
+completion(SCC) :-
+    % show_scc('Complete'),
+    '$tbl_component_status'(SCC, Status),
+    (   Status == active
+    ->  (   '$tbl_pop_worklist'(SCC, WorkList)
+        ->  tdebug(schedule, 'Complete WL ~p in ~p',
+                   [WorkList, SCC]),
+            completion_step(WorkList),
+            completion(SCC)
+        ;   tdebug(schedule, 'Completed ~p',
+                   [SCC]),
+            '$tbl_table_complete_all'(SCC)
+        )
+    ;   Status == merged
+    ->  tdebug(schedule, 'Aborted completion (SCC=~p)',
+               [scc(SCC)])
+    ;   true
+    ).
 
 completion_step(SourceTable) :-
     (   '$tbl_trienode'(Reserved),
@@ -250,7 +328,9 @@ completion_step(SourceTable) :-
                         Answer, ModeArgs,
                         Goal, Continuation, Wrapper, TargetTable),
         (   ModeArgs == Reserved
-        ->  Goal = Answer,
+        ->  tdebug(schedule, 'Resuming ~p, calling ~p with ~p',
+                   [Wrapper, Goal, Answer]),
+            Goal = Answer,
             delim(Wrapper, Continuation, TargetTable)
         ;   get_wrapper_no_mode_args(Goal, Answer, ModeArgs),
             get_wrapper_no_mode_args(Wrapper, WrapperNoModes, _),
@@ -333,6 +413,7 @@ wrappers(Name/Arity) -->
     { atom(Name), integer(Arity), Arity >= 0,
       !,
       functor(Head, Name, Arity),
+      check_undefined(Name/Arity),
       atom_concat(Name, ' tabled', WrapName),
       Head =.. [Name|Args],
       WrappedHead =.. [WrapName|Args],
@@ -350,6 +431,7 @@ wrappers(ModeDirectedSpec) -->
       !,
       functor(ModeDirectedSpec, Name, Arity),
       functor(Head, Name, Arity),
+      check_undefined(Name/Arity),
       atom_concat(Name, ' tabled', WrapName),
       Head =.. [Name|Args],
       WrappedHead =.. [WrapName|Args],
@@ -373,6 +455,20 @@ wrappers(TableSpec) -->
     { '$type_error'(table_desclaration, TableSpec)
     }.
 
+%!  check_undefined(+PI)
+%
+%   Verify the predicate has no clauses when the :- table is declared.
+%
+%   @tbd: future versions may rename the existing predicate.
+
+check_undefined(Name/Arity) :-
+    functor(Head, Name, Arity),
+    prolog_load_context(module, Module),
+    clause(Module:Head, _),
+    !,
+    '$permission_error'(table, procedure, Name/Arity).
+check_undefined(_).
+
 %!  mode_check(+Moded, -TestCode)
 %
 %   Enforce the output arguments of a  mode-directed tabled predicate to
@@ -564,6 +660,7 @@ sum(S0, S1, S) :- S is S0+S1.
 prolog:rename_predicate(M:Head0, M:Head) :-
     '$flushed_predicate'(M:'$tabled'(_)),
     call(M:'$tabled'(Head0)),
+    \+ current_prolog_flag(xref, true),
     !,
     rename_term(Head0, Head).
 
@@ -583,4 +680,5 @@ system:term_expansion((:- table(Preds)),
                         (:- multifile('$table_update'/4))
                       | Clauses
                       ]) :-
+    \+ current_prolog_flag(xref, true),
     phrase(wrappers(Preds), Clauses).
