@@ -79,27 +79,28 @@
 #define MSG_QLF_SECTION		 35
 #define MSG_QLF_BOOT		 36
 #define MSG_QLF_BOOT_READ	 37
-#define MSG_PROC_COUNT		 38
-#define MSG_CUT			 39
+#define MSG_QLF_LABEL		 38
+#define MSG_PROC_COUNT		 39
+#define MSG_CUT			 40
 
-#define MSG_QUEUE		 40
-#define MSG_QUEUE_WAIT		 41
-#define MSG_SIGNAL		 42
-#define MSG_COMP_VARS		 43
-#define MSG_DICT		 44
-#define MSG_PROF_CALLTREE	 45
-#define MSG_PROF_TICKS		 46
-#define MSG_INFERENCE_LIMIT	 47
-#define MSG_NSOLS		 48
-#define MSG_SRCFILE		 49
-#define MSG_DESTROY_MODULE	 50
-#define MSG_CREATE_MODULE	 51
-#define MSG_QUEUE_GC		 52
-#define MSG_ACYCLIC		 53
-#define MSG_OPERATOR		 54
-#define MSG_MUTEX_GC		 55
-#define MSG_REC_ATTVAR		 56
-#define MSG_TTY			 57
+#define MSG_QUEUE		 41
+#define MSG_QUEUE_WAIT		 42
+#define MSG_SIGNAL		 43
+#define MSG_COMP_VARS		 44
+#define MSG_DICT		 45
+#define MSG_PROF_CALLTREE	 46
+#define MSG_PROF_TICKS		 47
+#define MSG_INFERENCE_LIMIT	 48
+#define MSG_NSOLS		 40
+#define MSG_SRCFILE		 50
+#define MSG_DESTROY_MODULE	 51
+#define MSG_CREATE_MODULE	 52
+#define MSG_QUEUE_GC		 53
+#define MSG_ACYCLIC		 54
+#define MSG_OPERATOR		 55
+#define MSG_MUTEX_GC		 56
+#define MSG_REC_ATTVAR		 57
+#define MSG_TTY			 58
 
 #define MSG_READ_TOKEN		 60
 
@@ -177,6 +178,7 @@
 #define MSG_MODULE_REF		 260
 
 #define MSG_TABLING_WORK	 300
+#define MSG_TABLING_MODED	 301
 
 #define CHK_SECURE              1000
 #define CHK_HIGH_ARITY          1001
