@@ -48,7 +48,7 @@
 #define PLHOME       "c:/Program Files (x86)/swipl"
 #endif
 #define DEFSTARTUP   "swipl.ini"
-#else
+#else /*__WINDOWS__*/
 #include <config.h>
 #endif
 
@@ -73,11 +73,13 @@
 #define C_STATICLIBS ""
 #define C_CC	     "gcc"
 #define C_CFLAGS     ""
-#define C_PLLIB	     "-lswipl"
+#define C_PLLIB	     "-lswipl"		/* Or "libswipl.lib"? */
+#define C_LIBPLSO    "-lswipl"
 #define C_LDFLAGS    ""
-#endif
+#else
 #include <parms.h>			/* pick from the working dir */
 #endif
+#endif
 
 #define PL_KERNEL		1
 #include <inttypes.h>
@@ -242,7 +244,7 @@ gcc.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifndef __unix__
-#if defined(_AIX) || defined(__APPLE__) || defined(__unix) || defined(__BEOS__) || defined(__NetBSD__)
+#if defined(_AIX) || defined(__APPLE__) || defined(__unix) || defined(__BEOS__) || defined(__NetBSD__) || defined(__HAIKU__)
 #define __unix__ 1
 #endif
 #endif
@@ -930,12 +932,14 @@ Macros for environment frames (local stack frames)
 #define FR_SKIPPED		(0x0002) /* We have skipped on this frame */
 #define FR_MARKED		(0x0004) /* GC */
 #define FR_MARKED_PRED		(0x0008) /* GC predicates/clauses */
-#define FR_WATCHED		(0x0010) /* GUI debugger */
+#define FR_DEBUG		(0x0010) /* GUI debugger */
 #define FR_CATCHED		(0x0020) /* Frame caught an exception */
 #define FR_INBOX		(0x0040) /* Inside box (for REDO in built-in) */
 #define FR_CONTEXT		(0x0080) /* fr->context is set */
-#define FR_CLEANUP		(0x0100) /* setup_call_cleanup/4: marked for cleanup */
+#define FR_CLEANUP		(0x0100) /* setup_call_cleanup/4 */
 #define FR_INRESET		(0x0200) /* Continuations: inside reset/3 */
+#define FR_WATCHED (FR_CLEANUP|FR_DEBUG)
+
 #define FR_MAGIC_MASK		(0xfffff000)
 #define FR_MAGIC_MASK2		(0xffff0000)
 #define FR_MAGIC		(0x549d5000)
@@ -1878,10 +1882,26 @@ Temporary store/restore pointers to make them safe over GC/shift
 		 *	       SIGNALS		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+SWI-Prolog may be compiled without signal handling. Even in that case we
+still have signals that trigger Prolog   housekeeping  events. These are
+not bound to operating system signal handling though.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 #if HAVE_SIGNAL
 #define MAXSIGNAL		64	/* highest system signal number */
 #define SIG_PROLOG_OFFSET	32	/* Start of Prolog signals */
 
+#else /* HAVE_SIGNAL */
+
+#define MAXSIGNAL		32	/* highest system signal number */
+#define SIG_PROLOG_OFFSET	1	/* Start of Prolog signals */
+
+#endif /* HAVE_SIGNAL */
+
+#ifndef RETSIGTYPE
+#define RETSIGTYPE void
+#endif
 typedef RETSIGTYPE (*handler_t)(int);
 
 typedef struct
@@ -1890,18 +1910,17 @@ typedef struct
   predicate_t predicate;		/* Prolog handler */
   int	      flags;			/* PLSIG_*, defined in pl-setup.c */
 } sig_handler, *SigHandler;
-#endif /* HAVE_SIGNAL */
 
-#define SIG_EXCEPTION	  (SIG_PROLOG_OFFSET+0)
+
 #ifdef O_ATOMGC
-#define SIG_ATOM_GC	  (SIG_PROLOG_OFFSET+1)
+#define SIG_ATOM_GC	  (SIG_PROLOG_OFFSET+0)
 #endif
-#define SIG_GC		  (SIG_PROLOG_OFFSET+2)
+#define SIG_GC		  (SIG_PROLOG_OFFSET+1)
 #ifdef O_PLMT
-#define SIG_THREAD_SIGNAL (SIG_PROLOG_OFFSET+3)
+#define SIG_THREAD_SIGNAL (SIG_PROLOG_OFFSET+2)
 #endif
-#define SIG_CLAUSE_GC	  (SIG_PROLOG_OFFSET+4)
-#define SIG_PLABORT	  (SIG_PROLOG_OFFSET+5)
+#define SIG_CLAUSE_GC	  (SIG_PROLOG_OFFSET+3)
+#define SIG_PLABORT	  (SIG_PROLOG_OFFSET+4)
 
 
 		 /*******************************
@@ -2145,7 +2164,7 @@ typedef struct redir_context
 		*       READ WARNINGS           *
 		*********************************/
 
-#define ReadingSource (source_line_no > 0 && \
+#define ReadingSource (source_line_no >= 0 && \
 		       source_file_name != NULL_ATOM)
 
 
@@ -2195,7 +2214,7 @@ typedef struct
 #define MODULE_user	(GD->modules.user)
 #define MODULE_system	(GD->modules.system)
 #define MODULE_parse	(ReadingSource ? LD->modules.source \
-				       : MODULE_user)
+				       : LD->modules.typein)
 
 
 		/********************************
@@ -2318,7 +2337,7 @@ typedef struct debuginfo
 #define PLFLAG_AUTOLOAD		    0x00004000 /* do autoloading */
 #define PLFLAG_CHARCONVERSION	    0x00008000 /* do character-conversion */
 #define PLFLAG_LASTCALL		    0x00010000 /* Last call optimization enabled? */
-//				    0x00020000 /* not used */
+#define PLFLAG_PORTABLE_VMI	    0x00020000 /* Generate portable VMI code */
 #define PLFLAG_SIGNALS		    0x00040000 /* Handle signals */
 #define PLFLAG_DEBUGINFO	    0x00080000 /* generate debug info */
 #define PLFLAG_FILEERRORS	    0x00100000 /* Edinburgh file errors */
