@@ -88,10 +88,8 @@ setupProlog(void)
   initCharTypes();
   DEBUG(1, Sdprintf("foreign predicates ...\n"));
   initForeign();
-#if HAVE_SIGNAL
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
   initSignals();
-#endif
   DEBUG(1, Sdprintf("Stacks ...\n"));
   if ( !initPrologStacks(GD->options.stackLimit) )
     outOfCore();
@@ -199,9 +197,6 @@ they  define  signal handlers to be int functions.  This should be fixed
 some day.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if HAVE_SIGNAL
-#define HAVE_SIGNALS 1
-
 #define PLSIG_PREPARED 0x00010000	/* signal is prepared */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -216,6 +211,7 @@ static struct signame
   int	      flags;
 } signames[] =
 {
+#ifdef HAVE_SIGNAL
 #ifdef SIGHUP
   { SIGHUP,	"hup",    0},
 #endif
@@ -294,12 +290,12 @@ static struct signame
 #ifdef SIGPWR
   { SIGPWR,	"pwr",    0},
 #endif
+#endif /*HAVE_SIGNAL*/
 
 /* The signals below here are recorded as Prolog interrupts, but
    not supported by OS signals.  They start at offset 32.
 */
 
-  { SIG_EXCEPTION,     "prolog:exception",     0 },
 #ifdef SIG_ATOM_GC
   { SIG_ATOM_GC,       "prolog:atom_gc",       0 },
 #endif
@@ -532,7 +528,7 @@ dispatch_signal(int sig, int sync)
 
     PL_error(predname, arity, NULL, ERR_SIGNALLED, sig, signal_name(sig));
   } else if ( sh->handler )
-  {
+  { int ex_pending = (exception_term && !sync);
 #ifdef O_LIMIT_DEPTH
     uintptr_t olimit = depth_limit;
     depth_limit = DEPTH_NO_LIMIT;
@@ -546,11 +542,9 @@ dispatch_signal(int sig, int sync)
 	  Sdprintf("Handler %p finished (pending=0x%x,0x%x)\n",
 		   sh->handler, LD->signal.pending[0], LD->signal.pending[1]));
 
-    if ( exception_term && !sync )	/* handler: PL_raise_exception() */
-    { LD->signal.exception = PL_record(exception_term);
-      PL_raise(SIG_EXCEPTION);
-      exception_term = 0;
-    }
+    if ( !ex_pending && exception_term && !sync )	/* handler: PL_raise_exception() */
+      fatalError("Async exception handler for signal %s (%d) raised "
+		 "an exception", signal_name(sig), sig);
   }
 
   LD->signal.current = saved_current_signal;
@@ -600,7 +594,7 @@ set_sighandler(int sig, handler_t func)
     return old.sa_handler;
   else
     return SIG_DFL;
-#else
+#elif defined(HAVE_SIGNAL)
 #ifdef __WINDOWS__
   switch( sig )				/* Current Windows versions crash */
   { case SIGABRT:			/* when given a non-supported value */
@@ -613,16 +607,13 @@ set_sighandler(int sig, handler_t func)
     default:
       return SIG_IGN;
   }
-#endif
+#endif /*__WINDOWS__*/
   return signal(sig, func);
+#else
+  return NULL;
 #endif
 }
 
-#ifdef HAVE_SIGINFO_H
-#include <siginfo.h>
-#endif
-
-
 static SigHandler
 prepareSignal(int sig)
 { SigHandler sh = &GD->signals.handlers[sig-1];
@@ -662,6 +653,7 @@ hupHandler(int sig)
 #endif
 
 
+#ifdef HAVE_SIGNAL
 /* terminate_handler() is called on termination signals like SIGTERM.
    It runs hooks registered using PL_exit_hook() and then kills itself.
    The hooks are called with the exit status `3`.
@@ -693,26 +685,7 @@ initTerminationSignals(void)
   PL_signal(SIGQUIT, terminate_handler);
 #endif
 }
-
-static void
-sig_exception_handler(int sig)
-{ GET_LD
-  (void)sig;
-
-  if ( HAS_LD && LD->signal.exception )
-  { record_t ex = LD->signal.exception;
-
-    LD->signal.exception = 0;
-
-    PL_put_variable(exception_bin);
-    PL_recorded(ex, exception_bin);
-    PL_erase(ex);
-    exception_term = exception_bin;
-
-    DEBUG(CHK_SECURE, checkData(valTermRef(exception_term)));
-  }
-}
-
+#endif /*HAVE_SIGNAL*/
 
 static void
 agc_handler(int sig)
@@ -774,10 +747,12 @@ initSignals(void)
   /* This is general signal handling that is not strictly needed */
   if ( truePrologFlag(PLFLAG_SIGNALS) )
   { struct signame *sn = signames;
+#ifdef HAVE_SIGNAL
 #ifdef SIGPIPE
     set_sighandler(SIGPIPE, SIG_IGN);
 #endif
     initTerminationSignals();
+#endif /*HAVE_SIGNAL*/
     initBackTrace();
     for( ; sn->name; sn++)
     {
@@ -808,7 +783,6 @@ initSignals(void)
   /* these signals are not related to Unix signals and can thus */
   /* be enabled always */
 
-  PL_signal(SIG_EXCEPTION|PL_SIGSYNC,     sig_exception_handler);
   PL_signal(SIG_GC|PL_SIGSYNC,	          gc_handler);
   PL_signal(SIG_CLAUSE_GC|PL_SIGSYNC,     cgc_handler);
   PL_signal(SIG_PLABORT|PL_SIGSYNC,       abort_handler);
@@ -1024,12 +998,13 @@ PL_sigaction(int sig, pl_sigaction_t *act, pl_sigaction_t *old)
   return sig;
 }
 
+#ifndef SIG_DFL
+#define SIG_DFL (handler_t)-1
+#endif
 
 handler_t
 PL_signal(int sigandflags, handler_t func)
-{
-#ifdef HAVE_SIGNALS
-  pl_sigaction_t act = {0};
+{ pl_sigaction_t act = {0};
   pl_sigaction_t old;
 
   act.sa_cfunction = func;
@@ -1046,9 +1021,6 @@ PL_signal(int sigandflags, handler_t func)
   }
 
   return NULL;
-#else
-  return SIG_DFL;
-#endif
 }
 
 
@@ -1150,6 +1122,7 @@ endCritical__LD(ARG1_LD)
 }
 
 
+#ifdef HAVE_SIGNAL
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 on_signal(?SigNum, ?SigName, :OldHandler, :NewHandler)
 
@@ -1240,8 +1213,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
     { unprepareSignal(sign);
     } else if ( a == ATOM_throw )
     { sh = prepareSignal(sign);
-      set(sh, PLSIG_THROW);
-      clear(sh, PLSIG_SYNC);
+      set(sh, PLSIG_THROW|PLSIG_SYNC);
       sh->handler   = NULL;
       sh->predicate = NULL;
     } else
@@ -1791,9 +1763,11 @@ PRED_IMPL("$set_prolog_stack", 4, set_prolog_stack, 0)
 
 BeginPredDefs(setup)
   PRED_DEF("$set_prolog_stack",	  4, set_prolog_stack,	  0)
-  PRED_DEF("$on_signal",	  4, on_signal,		  0)
   PRED_DEF("trim_stacks",	  0, trim_stacks,	  0)
+#ifdef HAVE_SIGNAL
+  PRED_DEF("$on_signal",	  4, on_signal,		  0)
 #ifdef SIG_ALERT
   PRED_DEF("prolog_alert_signal", 2, prolog_alert_signal, 0)
 #endif
+#endif
 EndPredDefs
