@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2016, VU University Amsterdam
+    Copyright (c)  2017, VU University Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,6 +35,7 @@
 #ifndef _PL_TRIE_H
 #define _PL_TRIE_H
 #include "pl-indirect.h"
+#include "pl-copyterm.h"
 
 #define TRIE_MAGIC  0x4bcbcf87
 #define TRIE_CMAGIC 0x4bcbcf88
@@ -51,6 +52,8 @@ typedef struct try_children_any
 typedef struct trie_children_hashed
 { tn_node_type type;
   Table table;
+  unsigned int nvars;
+  size_t gsize;
 } trie_children_hashed;
 
 typedef struct trie_children_key
@@ -91,6 +94,7 @@ typedef struct trie
   struct
   { struct worklist *worklist;		/* tabling worklist */
     trie_node	    *variant;		/* node in variant trie */
+    fastheap_term   *skeleton;		/* Wrapper-Vars */
   } data;
 } trie;
 
@@ -101,7 +105,6 @@ typedef struct trie
 
 COMMON(void)	initTries(void);
 COMMON(trie *)	trie_create(void);
-COMMON(void)	trie_destroy(trie *trie);
 COMMON(void)	trie_empty(trie *trie);
 COMMON(void)	trie_clean(trie *trie);
 COMMON(void)	prune_node(trie *trie, trie_node *n);
@@ -113,5 +116,7 @@ COMMON(int)	trie_lookup(trie *trie, trie_node **nodep, Word k,
 COMMON(int)	trie_error(int rc, term_t culprit);
 COMMON(atom_t)	trie_symbol(trie *trie);
 COMMON(trie *)	symbol_trie(atom_t symbol);
+COMMON(int)	put_trie_value(term_t t, trie_node *node ARG_LD);
+COMMON(int)	set_trie_value(trie_node *node, term_t value ARG_LD);
 
 #endif /*_PL_TRIE_H*/
